from __future__ import absolute_import, division, print_function

from collections import Counter

import nltk
from nltk.corpus import stopwords

from dossier.models.features.stopwords import stopwords as dossier_stopwords


def sip_noun_phrases(tfidf, noun_phrases, limit=40):
    bow = tfidf[tfidf.id2word.doc2bow(noun_phrases)]
    return {tfidf.id2word[word]: count
            for word, count in Counter(dict(bow)).most_common(limit)}


def noun_phrases(text):
    '''Generate a bag of normalized stemmed noun phrases from ``text``.

    This is built around python's nltk library for getting Noun
    Phrases (NPs). This is all documented in the NLTK Book
    http://www.nltk.org/book/ch03.html and blog posts that cite the
    book.

    :rtype: list of phrase strings with spaces replaced by ``_``.
    '''
    ## from NLTK Book:
    sentence_re = r'''(?x)      # set flag to allow verbose regexps
          ([A-Z])(\.[A-Z])+\.?  # abbreviations, e.g. U.S.A.
        | \w+(-\w+)*            # words with optional internal hyphens
        | \$?\d+(\.\d+)?%?      # currency and percentages, e.g. $12.40, 82%
        | \.\.\.                # ellipsis
        | [][.,;"'?():-_`]      # these are separate tokens
    '''

    lemmatizer = nltk.WordNetLemmatizer()
    stemmer = nltk.stem.porter.PorterStemmer()

    ## From Su Nam Kim paper:
    ## http://www.comp.nus.edu.sg/~kanmy/papers/10.1007_s10579-012-9210-3.pdf
    grammar = r'''
        NBAR:
            {<NN.*|JJ>*<NN.*>}  # Nouns and Adjectives, terminated with Nouns

        NP:
            {<NBAR>}
            {<NBAR><IN><NBAR>}  # Above, connected with in/of/etc...
    '''
    if len(text.strip()) == 0:
        return []

    chunker = nltk.RegexpParser(grammar)

    toks = nltk.regexp_tokenize(text, sentence_re)
    postoks = nltk.tag.pos_tag(toks)

    #print postoks
    tree = chunker.parse(postoks)
    stops = stopwords.words('english')
    stops += dossier_stopwords()

    ## These next four functions are standard uses of NLTK illustrated by
    ## http://alexbowe.com/au-naturale/
    ## https://gist.github.com/alexbowe/879414
    def leaves(tree):
        '''Finds NP (nounphrase) leaf nodes of a chunk tree.'''
        for subtree in tree.subtrees(filter = lambda t: t.label()=='NP'):
            yield subtree.leaves()

    def normalise(word):
        '''Normalises words to lowercase and stems and lemmatizes it.'''
        word = word.lower()
        word = stemmer.stem_word(word)
        word = lemmatizer.lemmatize(word)
        return word

    def acceptable_word(word):
        '''Checks conditions for acceptable word: length, stopword.'''
        return 2 <= len(word) <= 40 and word.lower() not in stops

    def get_terms(tree):
        for leaf in leaves(tree):
            yield [normalise(w) for w,t in leaf if acceptable_word(w)]

    return ['_'.join(term) for term in get_terms(tree)]



from __future__ import absolute_import, division, print_function

from dossier.models.etl.ads import Ads
from dossier.models.etl.interface import ETL, add_sip_to_fc, html_to_fc
from dossier.models.etl.scrapy import Scrapy

__all__ = [
    'ETL', 'Ads', 'Scrapy',
    'add_sip_to_fc', 'html_to_fc',
]



from __future__ import absolute_import, division, print_function

import argparse
from itertools import chain, islice
import multiprocessing
from streamcorpus_pipeline._clean_visible import cleanse, make_clean_visible
import sys
import zlib

import cbor
from gensim import corpora, models
import happybase

import dossier.models.features as features
from dossier.web import streaming_sample
import kvlayer
import yakonfig


def status(*args, **kwargs):
    kwargs['end'] = ''
    args = list(args)
    args[0] = '\033[2K\r' + args[0]
    print(*args, **kwargs)
    sys.stdout.flush()


def batch_iter(n, iterable):
    iterable = iter(iterable)
    while True:
        yield chain([next(iterable)], islice(iterable, n-1))


def unpack_noun_phrases(row):
    body = cbor.loads(zlib.decompress(row['f:response.body']))
    body = make_clean_visible(body.encode('utf-8')).decode('utf-8')
    body = cleanse(body)
    return features.noun_phrases(body)


class App(yakonfig.cmd.ArgParseCmd):
    def __init__(self, *args, **kwargs):
        yakonfig.cmd.ArgParseCmd.__init__(self, *args, **kwargs)

    def args_tfidf(self, p):
        p.add_argument('--host', default='localhost')
        p.add_argument('--port', default=9090, type=int)
        p.add_argument('--table-prefix', default='')
        p.add_argument('--limit', default=100, type=int)
        p.add_argument('--batch-size', default=1000, type=int)
        p.add_argument('-p', '--processes',
                       default=multiprocessing.cpu_count(), type=int)
        p.add_argument('ids', metavar='INPUT_ROW_KEY_SAMPLE_FILE',
                       help='A file containing row keys to use for a sample.')
        p.add_argument('out', metavar='OUTPUT_TFIDF_MODEL_FILE',
                       help='The file path to write the tfidf model to.')

    def do_tfidf(self, args):
        conn = happybase.Connection(host=args.host, port=args.port,
                                    table_prefix=args.table_prefix)
        t = conn.table('artifact')
        corpus = []
        print('Extracting random sample...')
        sample = streaming_sample(open(args.ids), args.limit)

        print('Building corpus...')
        batches = batch_iter(args.batch_size, (s.strip() for s in sample))
        pool = multiprocessing.Pool(processes=args.processes)
        for i, batch in enumerate(batches, 1):
            rows = (row for _, row in t.rows(list(batch)))
            for noun_phrases in pool.imap(unpack_noun_phrases, rows):
                corpus.append(noun_phrases)
            status('%d of %d batches done' % (i, args.limit / args.batch_size))

        print('Computing model...')
        dictionary = corpora.Dictionary(corpus)
        bows = [dictionary.doc2bow(tokens) for tokens in corpus]
        tfidf = models.TfidfModel(bows, id2word=dictionary)
        tfidf.save(args.out)

    def args_ids(self, p):
        p.add_argument('--host', default='localhost')
        p.add_argument('--port', default=9090, type=int)
        p.add_argument('--table-prefix', default='')
        p.add_argument('--limit', default=100, type=int)

    def do_ids(self, args):
        conn = happybase.Connection(host=args.host, port=args.port,
                                    table_prefix=args.table_prefix)
        t = conn.table('artifact')
        hbase_filter = 'FirstKeyOnlyFilter() AND KeyOnlyFilter()'
        ids = islice(enumerate(t.scan(filter=hbase_filter)), args.limit)
        for i, (key, data) in ids:
            print(key)
            if i % 100000 == 0:
                print('%d keys received' % i, file=sys.stderr)


def main():
    p = argparse.ArgumentParser(
        description='Specific utilities for working with the ad corpus.')
    app = App()
    app.add_arguments(p)
    args = yakonfig.parse_args(p, [kvlayer, yakonfig])
    app.main(args)


if __name__ == '__main__':
    main()



from __future__ import absolute_import, division, print_function

from itertools import ifilter, imap, islice
import json
import re
import sys
import traceback
import urllib
from urlparse import urljoin, urlparse, urlunparse

import urlnorm

from dossier.fc import FeatureCollection, StringCounter
from dossier.models.etl.interface import ETL, html_to_fc, mk_content_id


class Scrapy(ETL):
    @staticmethod
    def detect_url_prefix(filelike):
        # This is a heuristic based on my observation of the data. This assumes
        # that all data is coming from a single domain and that all discovered
        # relative links have exactly the same prefix. I'm dubious. ---AG
        #
        # The heuristic is to find an absolute URL and a relative URL, then
        # discover the prefix.
        #
        # If one could not be detected `None` is returned.
        def find_prefix():
            for uabs in uabss:
                for urel in urels:
                    prefix = prefix_of(uabs, urel)
                    if prefix is not None:
                        return prefix
            return None

        def prefix_of(uabs, urel):
            if len(uabs) >= len(urel) and uabs[-len(urel):] == urel:
                return uabs[:-len(urel)]
            return None

        def add_url(url):
            if url is None:
                return
            url = urlunparse(urlparse(url)._replace(query=''))
            if re.search('^http', url):
                uabss.add(url)
            else:
                urels.add(url.lstrip('.'))

        uabss, urels = set(), set()
        for row in ifilter(None, imap(json_maybe_loads, filelike)):
            if row.get('_type') == 'ForumPostItem':
                add_url(row.get('thread_link'))
                add_url(row.get('author', {}).get('link'))
            elif row.get('_type') == 'CcaItem':
                add_url(row.get('url'))
            prefix = find_prefix()
            if prefix is not None:
                return prefix
        return None

    def __init__(self, filelike, url_prefix=None):
        self.rows = ifilter(None, imap(json_maybe_loads, filelike))
        self.url_prefix = url_prefix

    def cids_and_fcs(self, mapper, limit=5):
        posts = ifilter(lambda d: d.get('_type') == 'ForumPostItem', self.rows)
        return mapper(from_forum_post,
                      islice(imap(self.sanitize, posts), limit))

    def sanitize(self, post):
        for n in ['link', 'avatar']:
            if n not in post['author']:
                continue
            post['author'][n] = self.sanitize_url(post['author'][n])
        if 'thread_link' in post:
            post['thread_link'] = self.sanitize_url(post['thread_link'])
        return post

    def sanitize_url(self, url):
        url = normurl(url)
        if self.url_prefix is None:
            return url
        if re.search('^http', url):
            return url
        return normurl(urljoin(self.url_prefix, url))


def from_forum_post(row):
    cid = forum_post_id(row)
    try:
        fc = html_to_fc(row['content'].strip(),
                        url=row['thread_link'],
                        timestamp=forum_post_timestamp(row),
                        other_features=forum_post_features(row))
    except:
        fc = None
        print('Could not create FC for %s:' % cid, file=sys.stderr)
        print(traceback.format_exc())
    return cid, fc


def forum_post_features(row):
    fc = FeatureCollection()
    for k in row['author']:
        fc['post_author_' + k] = row['author'][k]

    if 'image_urls' in row:
        fc['image_url'] = StringCounter()
        for image_url in row['image_urls']:
            fc['image_url'][image_url] += 1

    others = ['parent_id', 'thread_id', 'thread_link', 'thread_name', 'title']
    for k in others:
        if k in row:
            fc['post_' + k] = uni(row[k])
    return fc


def forum_post_id(row):
    ticks = forum_post_timestamp(row)
    abs_url = row['thread_link']
    author = row['author'].get('username', 'unknown')
    return mk_content_id('|'.join(map(urlquote, [ticks, abs_url, author])))


def forum_post_timestamp(row):
    return str(int(row['created_at']) / 1000)


def urlquote(s):
    if isinstance(s, unicode):
        s = s.encode('utf-8')
    return urllib.quote(s, safe='~')


def normurl(url):
    try:
        return urlnorm.norm(url)
    except urlnorm.InvalidUrl:
        return urlnorm.norm_path('', url)


def uni(s):
    if isinstance(s, str):
        return unicode(s, 'utf-8')
    return s


def json_maybe_loads(s):
    try:
        d = json.loads(s)
    except:
        return None
    if 'thread_link' not in d:
        return None
    return d



'''
.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

Generate feature collections with your data
===========================================
This library ships with a command line program ``dossier.etl`` which
provides a rudimentary pipeline for transforming data from your database
to feature collections managed by :mod:`dossier.store`.

(Currently, ``dossier.etl`` is hard-coded to support a specific HBase
database, but it will be generalized as part of future work.)
'''
from __future__ import absolute_import, division, print_function

import abc
import time
import urllib

from streamcorpus_pipeline._clean_visible import cleanse, make_clean_visible
from streamcorpus_pipeline._clean_html import make_clean_html

from dossier.fc import FeatureCollection, StringCounter
import dossier.models.features as features


class ETL(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __init__(self):
        raise NotImplementedError

    @abc.abstractmethod
    def cids_and_fcs(self):
        raise NotImplementedError


def mk_content_id(key):
    return 'web|' + urllib.quote(key, safe='~')


def html_to_fc(html, url=None, timestamp=None, other_features=None):
    def add_feature(name, xs):
        if name not in fc:
            fc[name] = StringCounter()
        fc[name] += StringCounter(xs)

    html = uni(html)
    timestamp = timestamp or int(time.time() * 1000)
    other_features = other_features or {}
    url = url or ''

    clean_html = make_clean_html(html.encode('utf-8')).decode('utf-8')
    clean_vis = make_clean_visible(clean_html.encode('utf-8')).decode('utf-8')

    fc = FeatureCollection()
    fc[u'meta_raw'] = html
    fc[u'meta_clean_html'] = clean_html
    fc[u'meta_clean_visible'] = clean_vis
    fc[u'meta_timestamp'] = unicode(timestamp)
    fc[u'meta_url'] = uni(url)

    add_feature(u'phone', features.phones(clean_vis))
    add_feature(u'email', features.emails(clean_vis))
    add_feature(u'bowNP', features.noun_phrases(cleanse(clean_vis)))

    add_feature(u'image_url', features.image_urls(clean_html))
    add_feature(u'a_url', features.a_urls(clean_html))

    ## get parsed versions, extract usernames
    fc[u'img_url_path_dirs'] = features.path_dirs(fc[u'image_url'])
    fc[u'img_url_hostnames'] = features.host_names(fc[u'image_url'])
    fc[u'img_url_usernames'] = features.usernames(fc[u'image_url'])
    fc[u'a_url_path_dirs'] = features.path_dirs(fc[u'a_url'])
    fc[u'a_url_hostnames'] = features.host_names(fc[u'a_url'])
    fc[u'a_url_usernames'] = features.usernames(fc[u'a_url'])

    for feat_name, feat_val in other_features.iteritems():
        fc[feat_name] = feat_val

    return fc


def add_sip_to_fc(fc, tfidf, limit=40):
    if 'bowNP' not in fc:
        return
    sips = features.sip_noun_phrases(tfidf, fc['bowNP'].keys(), limit=limit)
    fc[u'bowNP_sip'] = StringCounter(sips)


def uni(s):
    if isinstance(s, str):
        try:
            return unicode(s, 'utf-8')
        except UnicodeDecodeError:
            return unicode(s, 'latin-1')
    return s



'''
Feature extraction
==================
``dossier.models.features`` provides some convenience functions for
feature extraction.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

.. autofunction:: noun_phrases
'''
from dossier.models.features.basic import emails, image_urls, phones, a_urls
from dossier.models.features.basic import host_names, path_dirs, usernames
from dossier.models.features.sip import noun_phrases, sip_noun_phrases
from dossier.models.features.stopwords import stopwords


__all__ = [
    'emails', 'image_urls', 'a_urls', 'phones',
    'noun_phrases', 'sip_noun_phrases',
    'stopwords', 'host_names', 'path_dirs', 'usernames'
]



'''
.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

``dossier.models`` provides search engines and a :mod:`dossier.web`
application for working with active learning.

.. automodule:: dossier.models.web.run
.. automodule:: dossier.models.pairwise
.. automodule:: dossier.models.features
.. automodule:: dossier.models.etl
'''
from dossier.models import features
from dossier.models.pairwise import PairwiseFeatureLearner, similar, dissimilar

__all__ = [
    'PairwiseFeatureLearner', 'similar', 'dissimilar',
    'features',
]



'''dossier.* namespace package has several subpackages, see
http://github.com/dossier for more info

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

'''
import pkg_resources
pkg_resources.declare_namespace(__name__)



from __future__ import absolute_import, division, print_function

import base64
import re

from dossier.fc import StringCounter


def subtopics(folders, folder_id, subfolder_id, ann_id=None):
    '''Yields an unordered generator of subtopics in a subfolder.

    Each item of the generator is a 4-tuple of ``content_id``,
    ``subtopic_id``, ``subtopic_type`` and ``data``. Subtopic type
    is one of the following Unicode strings: ``text``, ``image``
    or ``manual``. The type of ``data`` is dependent on the
    subtopic type. For ``image``, ``data`` is a ``(unicode, str)``,
    where the first element is the URL and the second element is
    the binary image data. For all other types, ``data`` is a
    ``unicode`` string.

    :param str folder_id: Folder id
    :param str subfolder_id: Subfolder id
    :param str ann_id: Username
    :rtype: generator of
            ``(content_id, subtopic_id, url, subtopic_type, data)``
    '''
    # This code will be changed soon. In essence, it implements the
    # convention established in SortingDesk for storing subtopic data.
    # Currently, subtopic data is stored in the FC that the data (i.e.,
    # image or snippet) came from. This is bad because it causes pretty
    # severe race conditions.
    #
    # Our current plan is to put each subtopic datum in its own FC. It will
    # require this code to make more FC fetches, but we should be able to
    # do it with one `store.get_many` call.
    items = folders.grouped_items(folder_id, subfolder_id, ann_id=ann_id)
    fcs = {cid: fc for cid, fc in folders.store.get_many(items.keys())}
    for cid, subids in items.iteritems():
        fc = fcs[cid]
        for subid in subids:
            try:
                data = typed_subtopic_data(fc, subid)
            except KeyError:
                # We have a dangling label folks!
                continue
            yield cid, subid, fc['meta_url'], subtopic_type(subid), data


def typed_subtopic_data(fc, subid):
    '''Returns typed subtopic data from an FC.'''
    # I don't think this code will change after we fix the data race bug. ---AG
    ty = subtopic_type(subid)
    data = get_unicode_feature(fc, subid)
    assert isinstance(data, unicode), \
        'data should be `unicode` but is %r' % type(data)
    if ty == 'image':
        img_data = get_unicode_feature(fc, subid + '|data')
        img = re.sub('^data:image/[a-zA-Z]+;base64,', '', img_data)
        img = base64.b64decode(img.decode('utf-8'))
        return data, img
    elif ty in ('text', 'manual'):
        return data
    raise ValueError('unrecognized subtopic type "%s"' % ty)


def get_unicode_feature(fc, feat_name):
    feat = fc[feat_name]
    if isinstance(feat, StringCounter) and len(feat) == 0:
        return u''
    return feat


def subtopic_type(subid):
    return subid.split('|')[1]



'''
Web services for active learning search engines
===============================================
This library ships with a command line application, ``dossier.models``,
that runs a web service. It is easy to run::

    dossier.models -c config.yaml

This runs it on ``localhost:8080``. A sample config file looks like:

.. code-block:: yaml

    dossier.store:
      feature_indexes: ['bowNP_sip', 'phone', 'region']

    dossier.models:
      tfidf_path: /path/to/tfidf/model.tfidf

    kvlayer:
      app_name: dossierstack
      namespace: myapp
      storage_type: redis
      storage_addresses: ['localhost:6379']

Where the ``tfidf_path`` corresponds to a TF-IDF model generated by
``gensim``. You can create one with the ``dossier.etl tfidf`` command.
(Note that this is optional!)

Once ``dossier.models`` is running, you can try the
Sorting Desk browser extension for Google Chrome:
https://chrome.google.com/webstore/detail/sorting-desk/ikcaehokdafneaiojndpmfbimilmlnid.
Once the extension is installed, you'll need to go to
its options and configure it to point to ``http://localhost:8080``.

Alternatively, you can follow the steps here to get a simple example
working on a sample data set:
https://github.com/dossier/dossier.models#running-a-simple-example.


Web service endpoints
---------------------
The web service has all of the same endpoints as :mod:`dossier.web`.
A couple of the endpoints are slightly enhanced to take advantage of
``dossier.models`` pairwise learning algorithm and feature extraction.
These enhanced endpoints are :func:`dossier.web.routes.v1_search` and
:func:`dossier.web.routes.v1_fc_put`. The ``v1_search`` endpoint is
enhanced with the addition of the :func:`dossier.models.similar` and
:func:`dossier.models.dissimilar` search engines. Additionally, the
``v1_fc_put`` accepts ``text/html`` content and will generate a feature
collection for you.


How the Sorting Desk browser extension works
--------------------------------------------
SortingDesk needs a ``dossier.models`` web server in order to function
properly. Namely, it uses ``dossier.models`` (and the underlying
DossierStack) to add/update feature collections, store ground truth
data as "labels," and run pairwise learning algorithms to rank relevant
search results. All of this information is saved to the underlying
database, so it is persistent across all user sessions.

``dossier.models`` also provides a folder/sub-folder organization
UI. Currently, this is built on top of Google Chrome's local storage.
Namely, it doesn't yet use the folder/subfolder web services described
in :mod:`dossier.web`. Therefore, folders/sub-folders don't yet persist
across multiple user sessions, but they will once we migrate off of
Chrome's local storage.
'''
from __future__ import absolute_import, division, print_function

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
import logging
import os.path as path
import urllib

from bs4 import BeautifulSoup
import bottle
try:
    from gensim import models
    TFIDF = True
except ImportError:
    TFIDF = False

from dossier.fc import StringCounter
from dossier.models import etl
from dossier.models.pairwise import dissimilar, similar
from dossier.models.report import ReportGenerator
import dossier.web as web
import dossier.web.config as config
import dossier.web.routes as routes
import yakonfig


logger = logging.getLogger(__name__)
web_static_path = path.join(path.split(__file__)[0], 'static')
bottle.TEMPLATE_PATH.insert(0, path.join(web_static_path, 'tpl'))


def add_routes(app):
    @app.get('/SortingQueue')
    def example_sortingqueue():
        return bottle.template('example-sortingqueue.html')

    @app.get('/SortingDesk')
    def example_sortingdesk():
        return bottle.template('example-sortingdesk.html')

    @app.get('/static/<name:path>')
    def v1_static(name):
        return bottle.static_file(name, root=web_static_path)

    @app.get('/dossier/v1/folder/<fid>/report')
    def v1_folder_report(request, response, folders, store, fid):
        response.headers['Content-Disposition'] = \
            'attachment; filename="report-%s.xlsx"' % fid
        response.headers['Content-Type'] = \
            'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
        user = request.query.get('annotator_id')

        gen = ReportGenerator(folders, web.Folders.id_to_name(fid), user=user)
        body = StringIO()
        gen.run(body)
        return body.getvalue()

    @app.put('/dossier/v1/feature-collection/<cid>', json=True)
    def v1_fc_put(request, response, store, tfidf, cid):
        '''Store a single feature collection.

        The route for this endpoint is:
        ``PUT /dossier/v1/feature-collections/<content_id>``.

        ``content_id`` is the id to associate with the given feature
        collection. The feature collection should be in the request
        body serialized as JSON.

        Alternatively, if the request's ``Content-type`` is
        ``text/html``, then a feature collection is generated from the
        HTML. The generated feature collection is then returned as a
        JSON payload.

        This endpoint returns status ``201`` upon successful
        storage otherwise. An existing feature collection with id
        ``content_id`` is overwritten.
        '''
        tfidf = tfidf or None
        if request.headers.get('content-type', '').startswith('text/html'):
            url = urllib.unquote(cid.split('|', 1)[1])
            fc = create_fc_from_html(url, request.body.read(), tfidf=tfidf)
            logger.info('created FC for "%r": %r', cid, fc)
            store.put([(cid, fc)])
            return routes.fc_to_json(fc)
        else:
            return routes.v1_fc_put(request, response, lambda x: x, store, cid)


def create_fc_from_html(url, html, tfidf=None):
    soup = BeautifulSoup(unicode(html, 'utf-8'))
    title = soup_get(soup, 'title', lambda v: v.get_text())
    body = soup_get(soup, 'body', lambda v: v.prettify())
    fc = etl.html_to_fc(body, url=url, other_features={
        u'title': title,
        u'titleBow': StringCounter(title.split()),
    })
    if fc is None:
        return None
    if tfidf is not None:
        etl.add_sip_to_fc(fc, tfidf)
    return fc


def soup_get(soup, sel, cont):
    v = soup.find(sel)
    if v is None:
        return u''
    else:
        return cont(v)


def same_subfolder(store, label_store):
    '''Filter out results in the same subfolder.'''
    folders = web.Folders(store, label_store)
    def init_filter(query_content_id):
        subfolders = folders.parent_subfolders(query_content_id)
        cids = set()
        for folder_id, subfolder_id in subfolders:
            for cid, subid in folders.items(folder_id, subfolder_id):
                cids.add(cid)

                # Also add directly connected labels too.
                for lab in label_store.directly_connected((cid, subid)):
                    cids.add(lab.other(cid))

        def p((content_id, fc)):
            return content_id not in cids
        return p
    return init_filter


def get_application():
    engines = {
        'dissimilar': dissimilar,
        'similar': similar,
        'random': web.engine_random,
        'index_scan': web.engine_index_scan,
    }
    args, application = web.get_application(
        routes=[add_routes], search_engines=engines,
        filter_preds={'already_labeled': same_subfolder})

    tfidf_model = False
    if TFIDF:
        try:
            conf = yakonfig.get_global_config('dossier.models')
            tfidf_path = conf['tfidf_path']
            tfidf_model = models.TfidfModel.load(tfidf_path)
        except KeyError:
            pass
    routes.app.install(
        config.create_injector('tfidf', lambda: tfidf_model))
    return args, application


def main():
    args, application = get_application()
    web.run_with_argv(args, application)


if __name__ == '__main__':
    main()



from __future__ import absolute_import, division, print_function

try:
    from gensim import corpora, models
    TFIDF = True
except ImportError:
    TFIDF = False
import pytest

from dossier.models.features import sip
from dossier.models.tests import kvl, store


@pytest.fixture
def tfidf():
    if not TFIDF:
        return
    doc1 = u'Andrew likes Diet Pepsi.'
    doc2 = u'Andrew knows the muffin man.'
    doc3 = u'Andrew lives near the muffin man on Shirley Lane.'
    corpus = map(sip.noun_phrases, [doc1, doc2, doc3])
    dictionary = corpora.Dictionary(corpus)
    bows = [dictionary.doc2bow(tokens) for tokens in corpus]
    return models.TfidfModel(bows, id2word=dictionary)


# def test_fc_generator(tfidf):
    # print(tfidf)
    # print(tfidf[u'Andrew likes Diet Pepsi'])
    # assert False



'''Exercises the pairwise active learning model.

Note that these tests aren't really machine learning tests. They
are emphatically unit tests. Therefore, when these tests pass, one
should not use that as evidence to conclude that learning is
implemented correctly. The tests only check whether the output of
functions matches this programmer's expectation.
'''
from __future__ import absolute_import, division, print_function

import pytest

from dossier.fc import FeatureCollection
from dossier.label import Label, CorefValue
from dossier.web import Folders

from dossier.models import PairwiseFeatureLearner
import dossier.models.pairwise as mod_pairwise
from dossier.models.tests import kvl, store, label_store


@pytest.fixture
def pairwise(store, label_store):
    store.put([('q', counter_fc({'fubar': 5, 'hi': 1}))])
    return PairwiseFeatureLearner(store, label_store, 'q')


def counter_fc(bow):
    return FeatureCollection({u'feature': bow})


def pos_label(id1, id2):
    # Don't care about annotators or subtopics.
    return Label(id1, id2, '', CorefValue.Positive)


def neg_label(id1, id2):
    return Label(id1, id2, '', CorefValue.Negative)


def test_subtopic_labels(store, label_store):
    def lab(cid1, sid1, cid2, sid2, neg=False):
        coref_val = CorefValue.Negative if neg else CorefValue.Positive
        return Label(cid1, cid2, 'unknown', coref_val, sid1, sid2)
    nlab = lambda a, b, c, d: lab(a, b, c, d, neg=True)
    def has_label(haystack, needle):
        return any(lab.same_subject_as(needle) and lab.value == needle.value
                   for lab in haystack)

    folders = Folders(store, label_store)
    folders.add_folder('top')
    folders.add_item('top', 'foo', 'a', 'ax')
    folders.add_item('top', 'foo', 'b', 'bx')
    folders.add_item('top', 'bar', 'c', 'cx')
    folders.add_item('top', 'bar', 'd', 'dx')
    folders.add_folder('other')
    folders.add_item('other', 'baz', 'e', 'ex')

    # Simulates a negative label from a search result.
    label_store.put(nlab('a', 'ax', 'r', None))
    label_store.put(nlab('b', 'bx', 'r', None))

    pairwise = PairwiseFeatureLearner(store, label_store, 'a', 'ax')

    labels = pairwise.infer_subtopic_labels()
    print('\n'.join(map(repr, labels)))

    assert has_label(labels, nlab('a', 'ax', 'e', 'ex'))
    assert has_label(labels, nlab('a', 'ax', 'c', 'cx'))
    assert has_label(labels, nlab('a', 'ax', 'd', 'dx'))
    assert has_label(labels, lab('a', 'ax', 'b', 'bx'))
    assert has_label(labels, nlab('a', 'ax', 'r', None))
    assert has_label(labels, nlab('b', 'bx', 'r', None))
    assert len(labels) == 8  # 2 for folders, 2 for results, 4 inferred


def test_canopy(pairwise):
    '''Make sure canopies are using indexes correctly.'''
    pairwise.store.put([
        ('abc', counter_fc({'fubar': 1})),
        ('xyz', counter_fc({'foo': 1})),
        ('mno', counter_fc({'hi': 10})),
    ])
    pairwise.query_fc = pairwise.store.get(pairwise.query_content_id)
    canopy = pairwise.canopy()
    # N.B. Doesn't include the query!
    assert {cid for cid, _ in canopy} == {'abc', 'mno'}


def test_canopy_limit(pairwise):
    '''Make sure canopies are using indexes correctly.'''
    pairwise.store.put([
        ('abc', counter_fc({'fubar': 1})),
        ('xyz', counter_fc({'foo': 1})),
        ('mno', counter_fc({'hi': 10})),
        ('def', counter_fc({'hi': 10})),
        ('ghi', counter_fc({'hi': 10})),
    ])
    pairwise.query_fc = pairwise.store.get(pairwise.query_content_id)
    canopy = pairwise.canopy()
    assert len(canopy) == 4

    limited = pairwise.canopy(limit=2)
    assert len(limited) == 2


def test_labels(pairwise):
    '''Make sure we get the right labels.

    The "right" labels means: label expansion on the connected
    component of the query.

    This doesn't test negative label inference.
    '''
    pairwise.label_store.put(pos_label('q', 'a'))
    pairwise.label_store.put(pos_label('a', 'b'))
    pairwise.label_store.put(pos_label('y', 'z'))

    fs = lambda *args: frozenset(args)
    got = {fs(lab.content_id1, lab.content_id2)
           for lab in pairwise.labels_from_query()}
    assert got == {
        fs('q', 'a'), fs('a', 'b'),
        fs('q', 'b'),  # <-- label expansion!
    }


def test_label_limit(pairwise):
    '''Make sure label limiting works.'''
    pairwise.label_store.put(pos_label('q', 'a'))
    pairwise.label_store.put(pos_label('a', 'b'))
    pairwise.label_store.put(neg_label('q', 'y'))
    pairwise.label_store.put(neg_label('y', 'z'))

    fs = lambda *args: frozenset(args)
    got = {fs(lab.content_id1, lab.content_id2)
           for lab in pairwise.labels_from_query(limit=2)}
    assert len(got) <= 2 * 2


def test_label_indexing():
    '''Make sure labels get converted to indices correctly.'''
    content_objs = [
        ('a', counter_fc({})),
        ('b', counter_fc({})),
        ('c', counter_fc({})),
    ]
    labels = [
        pos_label('a', 'c'),
        pos_label('c', 'b'),
        neg_label('b', 'a'),
    ]
    indiced = mod_pairwise.labels_to_indexed_coref_values(content_objs, labels)
    assert indiced == [
        # ({-1, 1}, index1, index2)
        (1, 0, 2),
        (1, 1, 2),
        (-1, 0, 1),
    ]


def test_vectorizable_features():
    '''Make sure we only do learning on the right features.

    The "right" features means features that can be vectorized
    by sklearn. Translation: they must be instances of
    collections.Mapping.
    '''
    fc = FeatureCollection({
        u'yes': {'fubar': 1},
        u'no': u'haha',
    })
    got = mod_pairwise.vectorizable_features([fc])
    assert got == ['yes']


def test_dissimilarities():
    '''Make sure computing dissimilarities works like we expect.

    This is probably too rigid.
    '''
    from numpy.testing import assert_approx_equal
    from scipy.spatial.distance import cosine

    fcs = [
        counter_fc({'a': 1, 'b': 2}),  # 0
        counter_fc({'a': 2, 'b': 3}),  # 1
        counter_fc({'a': 20, 'b': 0})  # 2
    ]
    dis = mod_pairwise.dissimilarities([u'feature'], fcs)[u'feature']

    got = [dis[0][1], dis[0][2], dis[1][2]]
    exp = [
        1 - cosine([1, 2], [2, 3]),
        1 - cosine([1, 2], [20, 0]),
        1 - cosine([2, 3], [20, 0]),
    ]
    [assert_approx_equal(got[i], exp[i]) for i in xrange(len(got))]


def interesting_training_data():
    '''Returns ([(content_id, FeatureCollection)], [Label])'''
    return [
        ('a', counter_fc({'ignored': 1, 'x': 50, 'y': 0})),
        ('b', counter_fc({'x': 0, 'y': 100})),
        ('c', counter_fc({'x': 0, 'y': 101})),
        ('d', counter_fc({'x': 0, 'y': 101})),
        ('e', counter_fc({'x': 0, 'y': 101})),
        ('f', counter_fc({'x': 0, 'y': 101})),
        ('g', counter_fc({'x': 0, 'y': 101})),
    ], [
        neg_label('a', 'c'),
        neg_label('a', 'b'),
        pos_label('b', 'c'),
        pos_label('b', 'd'),
        pos_label('b', 'e'),
        pos_label('b', 'f'),
        pos_label('b', 'g'),
    ]

def test_training(pairwise):
    '''Tests that training does something.'''
    from numpy.testing import assert_approx_equal

    content_objs, labels = interesting_training_data()
    indiced = mod_pairwise.labels_to_indexed_coref_values(content_objs, labels)
    names, model, vec = pairwise.train(content_objs, indiced)

    assert names == vec.get_feature_names()
    assert_approx_equal(model.coef_[0][0], 0.61903921)
    assert model.classes_.tolist() == [-1, 1]


@pytest.mark.xfail
def test_classify(pairwise):
    '''Tests the whole kit and kaboodle.'''
    content_objs, labels = interesting_training_data()
    pairwise.store.put(content_objs)
    for lab in labels:
        pairwise.label_store.put(lab)

    # Overwrite the query so we get some hits.
    pairwise.store.put([('q', counter_fc({'x': 5, 'y': 90}))])

    # And assign some labels to the query.
    # (label expansion should fill in the rest)
    pairwise.label_store.put(neg_label('q', 'a'))
    pairwise.label_store.put(pos_label('q', 'b'))

    candidate_probs = pairwise.probabilities()
    id_to_prob = {cid: p for (cid, _), p in candidate_probs}

    # This is a little silly, but it at least makes sure that an obvious
    # negative example is ranked below other obvious positive examples.
    assert all([id_to_prob['a'] <= id_to_prob[cid] for cid in id_to_prob])


def test_search_engine(store, label_store):
    '''Pretty much the same as classify, but for the search engine.'''
    content_objs, labels = interesting_training_data()
    store.put(content_objs)
    for lab in labels:
        label_store.put(lab)

    # Overwrite the query so we get some hits.
    store.put([('q', counter_fc({'x': 5, 'y': 90}))])

    # And assign some labels to the query.
    # (label expansion should fill in the rest)
    label_store.put(pos_label('q', 'b'))
    label_store.put(neg_label('q', 'a'))
    # These can be removed in negative label inference returns. ---AG
    label_store.put(neg_label('q', 'b'))
    label_store.put(neg_label('q', 'c'))

    engine = mod_pairwise.similar(store, label_store)
    results = engine('q', lambda _: True, 1)
    print(results)
    assert results['results'][0][0] != 'a'


def test_search_engine_limit(store, label_store):
    '''Pretty much the same as classify, but for the search engine.'''
    content_objs, labels = interesting_training_data()
    store.put(content_objs)
    for lab in labels:
        label_store.put(lab)

    # Overwrite the query so we get some hits.
    store.put([('q', counter_fc({'x': 5, 'y': 90}))])

    # And assign some labels to the query.
    # (label expansion should fill in the rest)
    label_store.put(neg_label('q', 'a'))
    label_store.put(pos_label('q', 'b'))

    engine = mod_pairwise.similar(store, label_store)
    results = engine('q', lambda _: True, 2)
    assert len(results['results']) == 2


def test_no_labels(store, label_store):
    '''Make sure the learner can handle zero labels.'''
    # Overwrite the query so we get some hits.
    store.put([('q', counter_fc({'x': 5, 'y': 90}))])
    engine = mod_pairwise.similar(store, label_store)
    results = engine('q', lambda _: True, 1)
    assert len(results['results']) == 0


def test_only_positive_labels(store, label_store):
    '''Make sure the learner can handle one class of labels.'''
    content_objs, labels = interesting_training_data()
    store.put(content_objs)
    for lab in labels:
        if lab.value == CorefValue.Positive:
            label_store.put(lab)

    # Overwrite the query so we get some hits.
    store.put([('q', counter_fc({'x': 5, 'y': 90}))])

    # And assign a label to the query.
    label_store.put(pos_label('q', 'b'))

    engine = mod_pairwise.similar(store, label_store)
    results = engine('q', lambda _: True, 100)
    # The search engine will fall back to a plain index scan if it has
    # insufficient training data.
    assert len(results['results']) >= 0


def test_only_negative_labels(store, label_store):
    '''Make sure the learner can handle one class of labels.'''
    content_objs, labels = interesting_training_data()
    store.put(content_objs)
    for lab in labels:
        if lab.value == CorefValue.Negative:
            label_store.put(lab)

    # Overwrite the query so we get some hits.
    store.put([('q', counter_fc({'x': 5, 'y': 90}))])

    # And assign a label to the query.
    label_store.put(neg_label('q', 'a'))

    engine = mod_pairwise.similar(store, label_store)
    results = engine('q', lambda _: True, 100)
    # The search engine will fall back to a plain index scan if it has
    # insufficient training data.
    assert len(results['results']) >= 0



import os

from setuptools import setup, find_packages

from version import get_git_version
VERSION, SOURCE_LABEL = get_git_version()
PROJECT = 'dossier.models'
AUTHOR = 'Diffeo, Inc.'
AUTHOR_EMAIL = 'support@diffeo.com'
URL = 'http://github.com/dossier/dossier.models'
DESC = 'Active learning models'


def read_file(file_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    return open(file_path).read()


setup(
    name=PROJECT,
    version=VERSION,
    description=DESC,
    license='MIT',
    long_description=read_file('README.md'),
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
    ],
    install_requires=[
        'beautifulsoup4',
        'dossier.fc >= 0.1.4',
        'dossier.label >= 0.1.5',
        'dossier.web >= 0.7.0',
        'gensim',
        'happybase',
        'joblib',
        'many_stop_words',
        'nltk',
        'numpy',
        'regex',
        'requests',
        'scipy',
        'scikit-learn',
        'streamcorpus-pipeline>=0.7.2',
        'pytest',
        'pytest-diffeo >= 0.1.4',
        'urlnorm >= 1.1.3',
        'xlsxwriter',
        'Pillow'
    ],
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'dossier.models = dossier.models.web.__main__:main',
            'dossier.models.soft_selectors = dossier.models.soft_selectors:main',
            'dossier.etl = dossier.models.etl:main',
        ],
    },
)
